/*
 *   -- micMAGMA (version 0.1) --
 *      Univ. of Tennessee, Knoxville
 *      Univ. of California, Berkeley
 *      Univ. of Colorado, Denver
 *      April 2012
 *
 * @author Mark Gates
 */

// defines stub functions, if no GPU implementation is available.
#include <stdio.h>
#include <stdlib.h>
#include <map>
#include <iostream>
#include <mkl.h>

#include "magma.h"
#include "magma_mic_bindings.h"


// ========================================
// initialization
magma_err_t
magma_init( void )
{
    // Use node 1 by default
    int node = 1;

//    int req_pn = 11, con_pn;
    int req_pn = 2049, con_pn;
    struct scif_portID portID;

    portID.node = node;
//    portID.port = 10;
    portID.port = 2050;

    if ((gEpd = scif_open()) < 0) {
      printf("scif_open failed with error %d\n", errno);
      return errno;
    }

    if ((con_pn = scif_bind(gEpd, req_pn)) < 0) {
      printf("scif_bind failed with error %d\n", errno);
      return errno;
    }

 retry:
    if ((scif_connect(gEpd, &portID)) != 0) {
      if (ECONNREFUSED == errno) {
        printf("scif_connect failed with error %d retrying\n", errno);
        goto retry;
      }
      printf("scif_connect failed with error %d\n", errno);
      return errno;
    }
    printf("scif_connect success\n");

    return MAGMA_SUCCESS;
}

// --------------------
magma_err_t
magma_finalize( void )
{
    // Close node 1 by default (has to be the same as in magma_init)
    //int node =1;

    int err = MAGMA_SUCCESS;
    int control_msg = CLOSE_MIC;

    if ((err = scif_send(gEpd, &control_msg, sizeof(control_msg), 1)) <= 0) {
      err = errno;
      printf("scif_send failed with err %d\n", errno);
      fflush(stdout);
    }

    close(gEpd);
    return err;
}


// ========================================
// memory allocation on the MIC
magma_err_t
magma_malloc( magma_ptr* ptrPtr, size_t nbyte )
{
    int err = 0;
    int control_msg = ALLOCATE_MEM;
    
    // make nbyte multiple of PAGE_SIZE
    if (nbyte % PAGE_SIZE != 0) {
        nbyte = (size_t)((nbyte / PAGE_SIZE) + 1) * PAGE_SIZE;
    }
    
    if (scif_send(gEpd, &control_msg, sizeof(control_msg), 1) <= 0) {
        err = errno;
        printf("magma_malloc: scif_send failed with err %d\n", errno);
        fflush(stdout);
    }
    
    magma_mic_mem_info mmm_info;
    mmm_info.nbyte = nbyte;
    if (scif_send(gEpd, &mmm_info, sizeof(mmm_info), 1) <= 0) {
        err = errno;
        printf("magma_malloc: scif_send failed with err %d\n", errno);
        fflush(stdout);
    }

    if (scif_recv(gEpd, &mmm_info, sizeof(mmm_info), 1) <= 0) {
        err = errno;
        printf("magma_malloc: scif_recv failed with err %d\n", errno);
        fflush(stdout);
    }
    
    (*ptrPtr) = (void *)(mmm_info.address);
    
    //save device pointer address, offset and number of bytes
    gDev_nbyte [(off_t)(*ptrPtr)] = nbyte;
    gDev_offset[(off_t)(*ptrPtr)] = mmm_info.offset;
    
    //printf("mic_malloc:address->0x%lx offset: 0x%lx byte: %d\n", (off_t)(*ptr), gDev_offset[(off_t)(*ptr)], gDev_nbyte[(off_t)(*ptr)]);
    
    return MAGMA_SUCCESS;
}

// --------------------
magma_err_t
magma_free( magma_ptr ptr )
{
    int err = 0;
    int control_msg = FREE_MEM;
    magma_mic_mem_info mmm_info;
    
    //send message for unregister
    if (scif_send(gEpd, &control_msg, sizeof(control_msg), 1) <= 0) {
        err = errno;
        printf("scif_send failed with err %d\n", errno);
        fflush(stdout);
    }
    
    //send unregister information
    mmm_info.address = (off_t)ptr;
    mmm_info.offset  = gDev_offset[(off_t)ptr];
    mmm_info.nbyte   = gDev_nbyte [(off_t)ptr];
    
    //printf("mic_free:address->0x%lx offset: 0x%lx byte: %d\n", (off_t)(ptr), gDev_offset[(off_t)(ptr)], gDev_nbyte[(off_t)(ptr)]);
    if (scif_send(gEpd, &mmm_info, sizeof(mmm_info), 1) <= 0) {
        err = errno;
        printf("scif_send failed with err %d\n", errno);
        fflush(stdout);
    }
    if (scif_recv(gEpd, &mmm_info, sizeof(mmm_info), 1) <= 0) {
        err = errno;
        printf("magma_free: scif_recv failed with err %d\n", errno);
        fflush(stdout);
    }
    
    //remove from the database
    std::map<off_t, size_t>::iterator devnbyte_it;
    std::map<off_t, off_t>::iterator devoff_it;
    
    devnbyte_it = gDev_nbyte.find((off_t)ptr);
    gDev_nbyte.erase(devnbyte_it);
    devoff_it = gDev_offset.find((off_t)ptr);
    gDev_offset.erase(devoff_it);
    
    return MAGMA_SUCCESS;
}

// --------------------
// memory allocation on the host CPU
magma_err_t
magma_malloc_host( void** ptrPtr, size_t size )
{
  int err = 0;
  off_t offset, suggested_offset;
  off_t address;

  //make size multiple of PAGE_SIZE                                                                             
  if (size % PAGE_SIZE != 0) {
    size = ((size / PAGE_SIZE) + 1) * PAGE_SIZE;
  }

  posix_memalign(ptrPtr, PAGE_SIZE, size);
  if((*ptrPtr) == NULL)
    return MIC_FAILURE;

  suggested_offset = 0x0;
  address = (off_t)*ptrPtr;

  if((offset = scif_register(gEpd, *ptrPtr, size, suggested_offset,
                             SCIF_PROT_READ | SCIF_PROT_WRITE, 0)) < 0) {
    err = errno;
    printf("scif_register failed with err %d\n", errno);
    fflush(stdout);
    return MIC_FAILURE;
  }

  //printf("Host side: address: 0x%lx offset: 0x%lx\n", address, offset);                                   
  suggested_offset += offset;

  //save offset, base address size                                                                          
  gHost_nbyte[(off_t)(*ptrPtr)] = size;
  gHost_offset[(off_t)(*ptrPtr)] = offset;

  return MAGMA_SUCCESS;
}

// --------------------
// this actually frees memory.
magma_err_t
magma_free_host( void* ptr )
{
  int err;
  off_t nbyte, offset;

  nbyte  = gHost_nbyte [(off_t)ptr];
  offset = gHost_offset[(off_t)ptr];

  std::map<off_t, size_t  >::iterator hnbyte_it;
  std::map<off_t, off_t>::iterator hoff_it;

  //unregister the allocated memory                                                                         
  if ((err = scif_unregister(gEpd, offset, nbyte)) < 0) {
    err = errno;
    printf("scif_unregister failed with err %d\n", errno);
    fflush(stdout);
    return MIC_FAILURE;
  }

  //remove from the database                                                                                
  hnbyte_it = gHost_nbyte.find((off_t)ptr);
  gHost_nbyte.erase(hnbyte_it);
  hoff_it = gHost_offset.find((off_t)ptr);
  gHost_offset.erase(hoff_it);

  free(ptr);

  return MAGMA_SUCCESS;
}


// ========================================
// device & queue support
magma_err_t
magma_get_devices(
    magma_device_t* devices,
    magma_int_t     size,
    magma_int_t*    numPtr )
{
    *numPtr = 1;
    printf( "%s( ptr, %d, numPtr )\n", __func__, size );
    return MAGMA_SUCCESS;
}

magma_err_t
magma_queue_create( magma_device_t device, magma_queue_t* queuePtr )
{
    printf( "%s( %d, %d )\n", __func__, device, *queuePtr );
    return MAGMA_SUCCESS;
}

magma_err_t
magma_queue_destroy( magma_queue_t  queue )
{
    printf( "%s( %d )\n", __func__, queue );
    return MAGMA_SUCCESS;
}

// waits for all kernels and all communication to complete.
magma_err_t
magma_queue_sync( magma_queue_t queue )
{
    magma_mic_synchronize_kernels();
    magma_event_t event;
    magma_mic_fence_mark( &event );
    magma_mic_fence_wait(  event );
    return MAGMA_SUCCESS;
}

// waits for all kernels to complete.
magma_err_t
magma_kernel_sync()
{
    magma_mic_synchronize_kernels();
    return MAGMA_SUCCESS;
}


// ========================================
// event support
magma_err_t
magma_event_create( magma_event_t* eventPtr )
{
    // nothing to do.
    return MAGMA_SUCCESS;
}

magma_err_t
magma_event_destroy( magma_event_t event )
{
    // nothing to do.
    return MAGMA_SUCCESS;
}

magma_err_t
magma_event_record( magma_event_t event, magma_queue_t queue )
{
    // not yet implemented
    printf( "%s( %d )\n", __func__, event );
    return MAGMA_SUCCESS;
}

magma_err_t
magma_event_query( magma_event_t event )
{
    // not yet implemented
    printf( "%s( %d )\n", __func__, event );
    return MAGMA_SUCCESS;
}

magma_err_t
magma_event_sync( magma_event_t event )
{
    magma_mic_fence_wait( event );
    return MAGMA_SUCCESS;
}
