/*
 *   -- micMAGMA (version 0.1) --
 *      Univ. of Tennessee, Knoxville
 *      Univ. of California, Berkeley
 *      Univ. of Colorado, Denver
 *      April 2012
 *
 */

#include <mkl.h>
#include <sys/time.h>
#include <sys/mman.h>

#include "magma.h"
#include "magma_mic_bindings.h"
#include "magma_z_bindings.h"
#include "magma_c_bindings.h"
#include "magma_d_bindings.h"
#include "magma_s_bindings.h"
#include "../../testing/flops.h"

#include "magma_s_server.h"
#include "magma_d_server.h"
#include "magma_c_server.h"
#include "magma_z_server.h"

// Flops formula
#define MICFMULS_GEMM(m, n, k) ((m) * (n) * (k))
#define MICFADDS_GEMM(m, n, k) ((m) * (n) * (k))
#define MICFMULS_GEMV(m, n)   ((m) * (n))
#define MICFADDS_GEMV(m, n)   ((m) * (n))
#define MICFMULS_TRMM_L(m, n) (0.5 * (n) * (m) * ((m)+1))
#define MICFADDS_TRMM_L(m, n) (0.5 * (n) * (m) * ((m)-1))
#define MICFMULS_TRMM_R(m, n) (0.5 * (m) * (n) * ((n)+1))
#define MICFADDS_TRMM_R(m, n) (0.5 * (m) * (n) * ((n)-1))
#define MICFMULS_TRSM_L(m, n) (0.5 * (n) * (m) * ((m)+1))
#define MICFADDS_TRSM_L(m, n) (0.5 * (n) * (m) * ((m)-1))
#define MICFMULS_TRSM_R(m, n) (0.5 * (m) * (n) * ((n)+1))
#define MICFADDS_TRSM_R(m, n) (0.5 * (m) * (n) * ((n)-1))
#define MICFMULS_SYRK(k, n) (0.5 * (k) * (n) * ((n)+1))
#define MICFADDS_SYRK(k, n) (0.5 * (k) * (n) * ((n)+1))

#define MICFLOPS_GEMM(m, n, k) (MICFMULS_GEMM(m, n, k) + MICFADDS_GEMM(m, n, k))
#define MICFLOPS_GEMV(m, n)    (MICFMULS_GEMV(m, n) + MICFADDS_GEMV(m, n))
#define MICFLOPS_TRSM_L(m, n)  (MICFMULS_TRSM_L(m, n) + MICFADDS_TRSM_L(m, n))
#define MICFLOPS_TRSM_R(m, n)  (MICFMULS_TRSM_R(m, n) + MICFADDS_TRSM_R(m, n))

#define MICFLOPS_SYRK(k, n)    (MICFMULS_SYRK(k, n) + MICFADDS_SYRK(k, n))

/* //////////////////////////////////////////////////////////////////////////
   -- Return time in seconds since arbitrary point (e.g., unix epoch).
*/
#include "../../control/constants.cpp"


scif_epd_t mic_epd;

int accept_connection() {
	int epd;
	struct scif_portID portID;
	//int req_pn = 10, con_pn;
	int req_pn = 2050, con_pn;
	
	//portID.node = 1;
	//portID.port = 11;

	if ((epd = scif_open()) < 0) {
		printf("scif_open failed with error %d\n", errno);
		MIC_FAILURE;
	}

	if ((con_pn = scif_bind(epd, req_pn)) < 0) {
		printf("scif_bind failed with error %d\n", errno);
		MIC_FAILURE;
	}

	if ((scif_listen(epd, 5)) < 0) {
		printf("Listen ioctl failed with error %d\n", errno);
		MIC_FAILURE;
	}

	if (((scif_accept(epd, &portID, &mic_epd, SCIF_ACCEPT_SYNC)) < 0) && (errno != EAGAIN)) {
		printf("scif_accept failed with errno %d\n", errno);
		MIC_FAILURE;
	}
	printf("scif_accept complete with SCIF_ACCEPT_SYNC is set\n");
	return MAGMA_SUCCESS;
}

int close_connection() {

        printf("Close the scif driver\n");
        close(mic_epd);
	return MAGMA_SUCCESS;
}

int mic_send(void *msg, int len) {
	int err;
	if ((err = scif_send(mic_epd, msg, len, 1)) <= 0) {
		err = errno;
		printf("scif_recv failed with err %d\n", errno);
		fflush(stdout);
		return MIC_FAILURE;
	}
	return MAGMA_SUCCESS;
}

int mic_recv(void *msg, int len) {
	int err;
	if ((err = scif_recv(mic_epd, msg, len, 1)) <= 0) {
		err = errno;
		printf("scif_recv failed with err %d\n", errno);
		fflush(stdout);
		return MIC_FAILURE;
	}
	return MAGMA_SUCCESS;
}

int mic_fence_mark(int *mark) {
	int err;
	if ((err = scif_fence_mark(mic_epd, SCIF_FENCE_INIT_PEER, mark))) { 
		err = errno; 
		printf("scif_fence_mark failed with err %d\n", errno); 
		fflush(stdout); 
		return MIC_FAILURE;
	} 
	return MAGMA_SUCCESS;
}

int mic_fence_wait(int *mark) {
	int err;
	if ((err = scif_fence_wait(mic_epd, *mark))) { 
		err = errno; 
		printf("scif_fence_wait failed with err %d\n", errno); 
		fflush(stdout); 
		return MIC_FAILURE;
	} 

	return MAGMA_SUCCESS;
}


void mic_print(int M, int N, void *A, int lda) {
        for(int kk = 0; kk < N; kk++) {
                for(int rr = 0; rr < M; rr++)
                        printf("%f ", ((float *)A)[kk * lda + rr]);
                printf("\n");
        }
}


int mic_free(void *ptr, off_t offset, size_t nbyte) {
	int MPAGE_SIZE = 2 * 1024 * 1024;
	int err;
	//printf("mic_free: 0x%lx offset: 0x%lx nby: %ld\n", (off_t)ptr, offset, nbyte);
	if ((err = scif_unregister(mic_epd, offset, nbyte)) < 0) {
		err = errno;
		printf("mic_free: scif_unregister failed with err %d\n", errno);
		fflush(stdout);
		return MIC_FAILURE;
	}

	//free(ptr);
	err = munmap(ptr, nbyte + MPAGE_SIZE & ~(MPAGE_SIZE - 1));
	//if(err != 0) {
	//	printf("%d munmap failed: %d\n", EINVAL, err);
	//	return MIC_FAILURE;
	//}
	//printf("mic_free: 0x%lx offset: 0x%lx nby: %d\n", (off_t)ptr, offset, nbyte);
	return MAGMA_SUCCESS;
}


int mic_allocate(void **ptr, size_t nbyte, off_t *offset)
{
	int err = 0;
	off_t suggested_offset;
	off_t address;
 	
	(*ptr) = mmap( NULL, nbyte, PROT_READ | PROT_WRITE, MAP_ANONYMOUS | MAP_PRIVATE | MAP_HUGETLB, 0/*-1*/, 0 ); 	
	//err = posix_memalign(ptr, PAGE_SIZE, nbyte);
	if (*ptr == MAP_FAILED || *ptr == NULL) {
	//if (*ptr == NULL) {
		printf("mmap  failed %d\n", err);
		return MIC_FAILURE;
	}

	suggested_offset = 0x0;
	address = (off_t)*ptr;

	if((*offset = scif_register(mic_epd, *ptr, nbyte, suggested_offset,
				SCIF_PROT_READ | SCIF_PROT_WRITE, 0)) < 0) {
		err = errno;
		printf("scif_register failed with err %d\n", errno);
		fflush(stdout);
		return MIC_FAILURE;
	}
	
	//printf("mic_malloc: 0x%lx offset: 0x%lx nbyte: %ld pagesize: %ld\n", address, *offset, nbyte, (size_t)sysconf(_SC_PAGESIZE));
	suggested_offset += *offset;

	return MAGMA_SUCCESS;
}

void do_work() {
	int err = 0, task_ID, m, n, k;
	magma_side_t side;
        magma_trans_t transA, transB;

	double gflops, alpha, beta;
        double mic_time;

	magma_mic_mem_info mmm_info;

	void *buf_address;
	off_t offset = 0;
	
	size_t nbyte = 4096;

	while(1) {
		if (mic_recv(&task_ID, sizeof(task_ID)) == MIC_FAILURE) 
			return;

		switch(task_ID) {
			case magma_mic_SYNC:
				//printf("sync\n");
				if (mic_send(&task_ID, sizeof(task_ID)) == MIC_FAILURE) {
					printf("scif_send failed for STRMM\n");
					return;				
				}
				break;

			case magma_mic_SGEMM:
				//printf("gemm\n");
				magma_mic_sgemm_param sgemm_param;

				if (mic_recv(&sgemm_param, sizeof(sgemm_param)) == MIC_FAILURE) {	
					printf("scif_recv failed for SGEMM\n");
					return;	
				}

				mic_time = dsecnd();
				call_sgemm(sgemm_param);
				mic_time = dsecnd() - mic_time;

				m = sgemm_param.m;
				n = sgemm_param.n;
				k = sgemm_param.k;
				gflops = MICFLOPS_GEMM((double)m, (double)n, (double)k) * 1e-9;
				printf("sgemm: %d\t%d\t%d\t%g\t%g\t%g\n", m, n, k, mic_time, 
				       gflops, gflops/mic_time); 
				break;

			case magma_mic_DGEMM:
				//printf("gemm\n");
				magma_mic_dgemm_param dgemm_param;

				if (mic_recv(&dgemm_param, sizeof(dgemm_param)) == MIC_FAILURE) {	
					printf("scif_recv failed for DGEMM\n");
					return;	
				}

				mic_time = dsecnd();
				call_dgemm(dgemm_param);
				mic_time = dsecnd() - mic_time;;
				
				transA = dgemm_param.transa;
				transB = dgemm_param.transb;
				m = dgemm_param.m;
				n = dgemm_param.n;
				k = dgemm_param.k;
				alpha = dgemm_param.alpha;
				beta = dgemm_param.beta;
				gflops = MICFLOPS_GEMM((double)m, (double)n, (double)k) * 1e-9;
				printf("DGEMM: %4d\t%4d\t%4d\t%8.4lf sec\t%4.2lf GFlop/s\n", m, n, k, mic_time, gflops/mic_time); 
				break;

		        case magma_mic_SGEMV:
		                printf("gemv\n");  
			        magma_mic_sgemv_param sgemv_param;

				if (mic_recv(&sgemv_param, sizeof(sgemv_param)) == MIC_FAILURE) {
				  printf("scif_recv failed for SGEMV\n");
				  return;
				}
				
				mic_time = dsecnd();
				call_sgemv(sgemv_param);
				mic_time = dsecnd() - mic_time;

				m = sgemv_param.m;
				n = sgemv_param.n;
				gflops = MICFLOPS_GEMV((double)m, (double)n) * 1e-9;
				printf("sgemv: %d\t%d\t%g\t%g\t%g\n", m, n, mic_time,
				       gflops, gflops/mic_time);
				break;

		        case magma_mic_DGEMV:
		                printf("gemv\n");
			        magma_mic_dgemv_param dgemv_param;

				if (mic_recv(&dgemv_param, sizeof(dgemv_param)) == MIC_FAILURE) {
				  printf("scif_recv failed for DGEMV\n");
				  return;
				}

				mic_time = dsecnd();
				call_dgemv(dgemv_param);
				mic_time = dsecnd() - mic_time;;
				
				transA = dgemv_param.transa;
				m = dgemv_param.m;
				n = dgemv_param.n;
				alpha = dgemv_param.alpha;
				beta = dgemv_param.beta;
				gflops = MICFLOPS_GEMV((double)m, (double)n) * 1e-9;
				printf("DGEMV: %4d\t%4d\t%8.4lf sec\t%4.2lf GFlop/s\n", m, n, 
				       mic_time, gflops/mic_time);
				break;

			case magma_mic_STRSM:
				//printf("strsm\n");
				magma_mic_strsm_param strsm_param;
				if (mic_recv(&strsm_param, sizeof(strsm_param)) == MIC_FAILURE) {
					printf("scif_recv failed for STRSM\n");
					return;				
				}
				
				mic_time = dsecnd();  
				call_strsm(strsm_param);
				mic_time = dsecnd() - mic_time;

				side = strsm_param.side;
				m = strsm_param.m;
				n = strsm_param.n;

				if(side == 'L')
					gflops = MICFLOPS_TRSM_L((float)m, (float)n) * 1e-9;
				else
					gflops = MICFLOPS_TRSM_R((float)m, (float)n) * 1e-9;

				//printf("strsm: %c\t%d\t%d\t%f\t%f\t%f\n", side, m, n, mic_time, 
				  //     gflops, gflops/mic_time); 
				break;

			case magma_mic_DTRSM:
				//printf("dtrsm\n");
				magma_mic_dtrsm_param dtrsm_param;
				if (mic_recv(&dtrsm_param, sizeof(dtrsm_param)) == MIC_FAILURE) {
					printf("scif_recv failed for DTRSM\n");
					return;				
				}
				
				mic_time = dsecnd();
				call_dtrsm(dtrsm_param);
				mic_time = dsecnd() - mic_time;

				side = dtrsm_param.side;
				m = dtrsm_param.m;
				n = dtrsm_param.n;

				gflops = 0;
				if(lapack_const(side)[0] == 'L')
					gflops = MICFLOPS_TRSM_L((float)m, (float)n) * 1e-9;
				else if(lapack_const(side)[0] == 'R')
					gflops = MICFLOPS_TRSM_R((float)m, (float)n) * 1e-9;

				printf("TRSM:  %4d\t%4d\t\t%8.4lf sec\t%4.2lf Gflop/s\n", m, n, mic_time, gflops/mic_time); 
				break;

                        case magma_mic_STRMM:
				//printf("strmm\n");
                                magma_mic_strmm_param strmm_param;
                                if (mic_recv(&strmm_param, sizeof(strmm_param)) == MIC_FAILURE) {
                                        printf("scif_recv failed for STRSM\n");
                                        return;
                                }

                                mic_time = dsecnd();  
                                call_strmm(strmm_param);
                                mic_time = dsecnd() - mic_time;

                                side = strsm_param.side;
                                m = strsm_param.m;
                                n = strsm_param.n;

                                if(side == 'L')
                                        gflops = MICFLOPS_TRSM_L((float)m, (float)n) * 1e-9;
                                else
                                        gflops = MICFLOPS_TRSM_R((float)m, (float)n) * 1e-9;

                                //printf("strsm: %c\t%d\t%d\t%f\t%f\t%f\n", side, m, n, mic_time, 
				  //     gflops, gflops/mic_time); 
                                break;

                        case magma_mic_DTRMM:
				//printf("dtrmm\n");
                                magma_mic_dtrmm_param dtrmm_param;
                                if (mic_recv(&dtrmm_param, sizeof(dtrmm_param)) == MIC_FAILURE) {
                                        printf("scif_recv failed for DTRSM\n");
                                        return;
                                }

                                mic_time = dsecnd();  
                                call_dtrmm(dtrmm_param);
                                mic_time = dsecnd() - mic_time;

                                side = dtrsm_param.side;
                                m = dtrsm_param.m;
                                n = dtrsm_param.n;

                                //if(side == 'L')
                                //        gflops = MICFLOPS_TRSM_L((float)m, (float)n) * 1e-9;
                                //else
                                //        gflops = MICFLOPS_TRSM_R((float)m, (float)n) * 1e-9;

                                //printf("strsm: %s\t%d\t%d\t%g\n", lapack_const(side), m, n, gflops/mic_time); 
                                break;

			case magma_mic_SSYRK:
				//printf("ssyrk\n");
				magma_mic_ssyrk_param ssyrk_param;
				if (mic_recv(&ssyrk_param, sizeof(ssyrk_param)) == MIC_FAILURE) {
					printf("scif_recv failed for SSYRK\n");
					return;
				}

				mic_time = dsecnd();
				call_ssyrk(ssyrk_param);
				mic_time = dsecnd() - mic_time;

				n = ssyrk_param.n;
				k = ssyrk_param.k;
				gflops = MICFLOPS_SYRK((float)n, (float)k) * 1e-9;
				//printf("ssyrk: %d\t%d\t%f\t%f\t%f\n", n, k, mic_time, 
				  //     gflops, gflops/mic_time); 
				break;

			case magma_mic_DSYRK:
				//printf("dsyrk\n");
				magma_mic_dsyrk_param dsyrk_param;
				if (mic_recv(&dsyrk_param, sizeof(dsyrk_param)) == MIC_FAILURE) {
					printf("scif_recv failed for DSYRK\n");
					return;
				}

				mic_time = dsecnd();
				call_dsyrk(dsyrk_param);
				mic_time = dsecnd() - mic_time;

				n = dsyrk_param.n;
				k = dsyrk_param.k;
				gflops = MICFLOPS_SYRK((float)n, (float)k) * 1e-9;
				//printf("dsyrk: %d\t%d\t%g\n", n, k, gflops/mic_time); 
				break;

			case magma_mic_SLARFT:
				//printf("slarft\n");
				magma_mic_slarft_param slarft_param;
				if (mic_recv(&slarft_param, sizeof(slarft_param)) == MIC_FAILURE) {
					printf("scif_recv failed for SLARFT\n");
					return;
				}

				mic_time = dsecnd();
				call_slarft(slarft_param);
				mic_time = dsecnd() - mic_time;

				gflops = MICFLOPS_GEMM((float)m, (float)n, (float)k) * 1e-9;
				//printf("slarft: %d\t%d\t%d\t%f\t%f\t%f\n", m, n, k, mic_time,
				  //     gflops, gflops/mic_time); 
				break;

			case magma_mic_DLARFT:
				//printf("dlarft\n");
				magma_mic_dlarft_param dlarft_param;
				if (mic_recv(&dlarft_param, sizeof(dlarft_param)) == MIC_FAILURE) {
					printf("scif_recv failed for DLARFT\n");
					return;
				}

				mic_time = dsecnd();
				call_dlarft(dlarft_param);
				mic_time = dsecnd() - mic_time;

				gflops = MICFLOPS_GEMM((float)m, (float)n, (float)k) * 1e-9;
				//printf("slarft: %d\t%d\t%d\t%f\t%f\t%f\n", m, n, k, mic_time,
				  //     gflops, gflops/mic_time); 
				break;

			case magma_mic_SLARFB:
				//printf("slarfb\n");
				magma_mic_slarfb_param slarfb_param;
				if (mic_recv(&slarfb_param, sizeof(slarfb_param)) == MIC_FAILURE) {
					printf("scif_recv failed for SLARFB\n");
					return;
				}

				mic_time = dsecnd();
				call_slarfb(slarfb_param);
				mic_time = dsecnd();

				gflops = MICFLOPS_GEMM((float)m, (float)n, (float)k) * 1e-9;
				//printf("slarfb: %d\t%d\t%d\t%f\t%f\t%f\n", m, n, k, mic_time,
				  //     gflops, gflops/mic_time); 
				break;

			case magma_mic_DLARFB:
				//printf("dlarfb\n");
				magma_mic_dlarfb_param dlarfb_param;
				if (mic_recv(&dlarfb_param, sizeof(dlarfb_param)) == MIC_FAILURE) {
					printf("scif_recv failed for DLARFB\n");
					return;
				}

				mic_time = dsecnd();
				call_dlarfb(dlarfb_param);
				mic_time = dsecnd();

				gflops = MICFLOPS_GEMM((float)m, (float)n, (float)k) * 1e-9;
				//printf("slarfb: %d\t%d\t%d\t%f\t%f\t%f\n", m, n, k, mic_time,
				  //     gflops, gflops/mic_time); 
				break;

			case ALLOCATE_MEM:
				//printf("allocate mem\n");
				if(mic_recv(&mmm_info, sizeof(mmm_info)) == MIC_FAILURE) {
					printf("scif_recv failed for ALLOCATE_MEM\n");
					return;
				}

				//allocate
				if(mic_allocate(&buf_address, mmm_info.nbyte, &offset) == MIC_FAILURE)
					return;

				mmm_info.address = (off_t)(buf_address);
				mmm_info.offset = offset;

				if(mic_send(&mmm_info, sizeof(mmm_info)) == MIC_FAILURE) {
					printf("scif_send failed for ALLOCATE_MEM\n");
					return;				
				}
				break;

			case FREE_MEM:
				//printf("free mem\n");
				if (mic_recv(&mmm_info, sizeof(mmm_info)) == MIC_FAILURE) {
					printf("scif_recv failed for FREE_MEM\n");
					return;
				}

				buf_address = (void *)mmm_info.address;
				offset = mmm_info.offset;
				nbyte = mmm_info.nbyte;
				mic_free(buf_address, offset, nbyte);
				if(mic_send(&mmm_info, sizeof(mmm_info)) == MIC_FAILURE) {
					printf("scif_send failed for FREE_MEM\n");
					return;				
				}

				break;

                        case PRINT_dA:
				//printf("print mat\n");
				magma_mic_print_param print_param;
                                if (mic_recv(&print_param, sizeof(print_param)) == MIC_FAILURE) {
                                        printf("scif_recv failed for PRINT_dA\n");
					return;
                                }

                                mic_print(print_param.m, print_param.n, (void *)print_param.d_A, print_param.lda);
                                break;

		        case magma_mic_SLASWP:
				//printf("swap\n");
			        magma_mic_slaswp_param slaswp_param;
				if (mic_recv(&slaswp_param, sizeof(slaswp_param)) == MIC_FAILURE) {
				  printf("scif_recv failed for SLASWP\n");
				  return;
				}

				call_slaswp(slaswp_param);
				
				//if (mic_send(&task_ID, sizeof(task_ID)) == MIC_FAILURE) {
				//  printf("scif_send failed for SLASWP\n");
				//  return;
				//}
				break;

		        case magma_mic_DLASWP:
				//printf("swap\n");
			        magma_mic_dlaswp_param dlaswp_param;
				if (mic_recv(&dlaswp_param, sizeof(dlaswp_param)) == MIC_FAILURE) {
				  printf("scif_recv failed for DLASWP\n");
				  return;
				}

				call_dlaswp(dlaswp_param);
				
				//if (mic_send(&task_ID, sizeof(task_ID)) == MIC_FAILURE) {
				//  printf("scif_send failed for DLASWP\n");
				//  return;
				//}
				break;

		        case magma_mic_STRANSPOSE:
				//printf("transpose\n");
			        magma_mic_stranspose_param stranspose_param;
				if (mic_recv(&stranspose_param, sizeof(stranspose_param)) == MIC_FAILURE) {
				  printf("scif_recv failed for STRANSPOSE\n");
				  return;
				}

				call_stranspose(stranspose_param);

				//if (mic_send(&task_ID, sizeof(task_ID)) == MIC_FAILURE) {
				//  printf("scif_send failed for STRANSPOSE\n");
				//  return;
				//}
				break;

		        case magma_mic_DTRANSPOSE:
				//printf("transpose\n");
			        magma_mic_dtranspose_param dtranspose_param;
				if (mic_recv(&dtranspose_param, sizeof(dtranspose_param)) == MIC_FAILURE) {
				  printf("scif_recv failed for DTRANSPOSE\n");
				  return;
				}

				call_dtranspose(dtranspose_param);

				//if (mic_send(&task_ID, sizeof(task_ID)) == MIC_FAILURE) {
				//  printf("scif_send failed for DTRANSPOSE\n");
				//  return;
				//}
				break;
	
		        case CLOSE_MIC:
				//printf("close\n");
				//fclose(fp);
				return;
	
			default:
				printf("Unknown task\n");
		}

	}
}

int main(int argc, char **argv)
{
	if(accept_connection() != MAGMA_SUCCESS)
		return MIC_FAILURE;

        do_work();

	if(close_connection() != MAGMA_SUCCESS)
		return MIC_FAILURE;

	return 0;
}
